////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FFGL.h
//
// FreeFrame is an open-source cross-platform real-time video effects plugin system.
// It provides a framework for developing video effects plugins and hosts on Windows,
// Linux and Mac OSX.
//
// Copyright (c) 2018 www.freeframe.org
// All rights reserved.
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Redistribution and use in source and binary forms, with or without modification,
//	are permitted provided that the following conditions are met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in
//    the documentation and/or other materials provided with the
//    distribution.
//  * Neither the name of FreeFrame nor the names of its
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
//
//	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
//	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
//	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
//	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
//	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
//	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
//	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
//	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
//	OF THE POSSIBILITY OF SUCH DAMAGE.
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// First version, Marcus Clements (marcus@freeframe.org)
// www.freeframe.org
//
// FreeFrame 1.0 upgrade by Pete Warden
// www.petewarden.com
//
// FreeFrame 1.0 - 03 upgrade by Gualtiero Volpe
// Gualtiero.Volpe@poste.it
//
// #ifdef tweaks for FreeFrameGL upgrade by Trey Harrison
// www.harrisondigitalmedia.com
//
// FFGL 1.0 by Trey Harrison
// www.harrisondigitalmedia.com
//
// FFGL 2.0 by Menno Vink (menno@resolume.com)
// www.resolume.com
// -Removed the old FreeFrame functionality, and thus CPU processing is no longer supported.
// -Completed the FF_TYPE_OPTION parameter type. Plugins can add parameter elements which hosts
//  should use to show a dropdown styled parameter. When one of the options is selected the parameter
//  will get a value equal to that of the option's value.
// -Completed the FF_TYPE_BUFFER parameter type. When this parameter is used it should provide a usage indication
//  which the host can then use to fill that buffer with the requested data. This can be used by plugins to
//  access the host's global fft data for example.
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#pragma once

//////////////////////////////////////////////////////////////////////////////////////
// Includes
/////////////////////////////////////////////////////////////////////////////////////

//Include the appropriate OpenGL headers for the compiler

#if defined( _WIN32 )
#define WIN32_LEAN_AND_MEAN//Exclude rarely-used stuff from Windows headers
#define NOMINMAX           //No min/max macros.
#define _WINSOCKAPI_       //Prevent inclusion of winsock
#include <windows.h>
#include "../glsdk_0_5_2/glload/include/gl_4_1.h"
typedef unsigned __int16 FFUInt16;
typedef unsigned __int32 FFUInt32;
#else
#if defined( TARGET_OS_MAC )
#include <OpenGL/gl3.h>
#elif defined( __linux__ )
#include <GL/gl.h>
#else
#error define this for your OS
#endif

extern "C" {
#include <string.h>
#include <stdlib.h>
#include <stdint.h>

typedef uint16_t FFUInt16;
typedef uint32_t FFUInt32;
#endif

namespace ffgl
{

// Function codes
// Old enumerators that are no longer in use are tagged with comments ("No longer in use").
// This is to discourage using their values for new enumerators.
enum class FunctionCode : FFUInt32
{
	FF_GETINFO                       = 0,
	FF_INITIALISE                    = 1,// No longer in use
	FF_DEINITIALISE                  = 2,
	FF_PROCESSFRAME                  = 3,// No longer in use
	FF_GETNUMPARAMETERS              = 4,
	FF_GETPARAMETERNAME              = 5,
	FF_GETPARAMETERDEFAULT           = 6,
	FF_GETPARAMETERDISPLAY           = 7,
	FF_SETPARAMETER                  = 8,
	FF_GETPARAMETER                  = 9,
	FF_GETPLUGINCAPS                 = 10,
	FF_INSTANTIATE                   = 11,// No longer in use
	FF_DEINSTANTIATE                 = 12,// No longer in use
	FF_GETEXTENDEDINFO               = 13,
	FF_PROCESSFRAMECOPY              = 14,// No longer in use
	FF_GETPARAMETERTYPE              = 15,
	FF_GETINPUTSTATUS                = 16,
	FF_PROCESSOPENGL                 = 17,
	FF_INSTANTIATEGL                 = 18,
	FF_DEINSTANTIATEGL               = 19,
	FF_SETTIME                       = 20,
	FF_CONNECT                       = 21,
	FF_DISCONNECT                    = 22,
	FF_RESIZE                        = 23,
	FF_GETNUMPARAMETERELEMENTS       = 31,
	FF_INITIALISE_V2                 = 34,
	FF_GET_PARAMETER_ELEMENT_NAME    = 35,
	FF_GET_PARAMETER_ELEMENT_DEFAULT = 36,
	FF_SET_PARAMETER_ELEMENT_VALUE   = 37,
	FF_GETPARAMETERUSAGE             = 32,
	FF_GETPLUGINSHORTNAME            = 33,
	FF_SET_BEATINFO                  = 38,
	FF_SET_HOSTINFO                  = 39,
	FF_SET_SAMPLERATE                = 40,
	FF_GET_RANGE                     = 41,
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FreeFrame defines
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

typedef FFUInt32 FFResult;

enum SuccessFail : FFResult
{
	FF_SUCCESS = 0,
	FF_FAIL    = 0xFFFFFFFF
};

// Return values
enum BooleanValue : FFResult
{
	FF_TRUE  = 1,
	FF_FALSE = 0
};

enum Supported : FFUInt32
{
	FF_SUPPORTED   = 1,
	FF_UNSUPPORTED = 0
};

// Plugin types
enum class PluginType : FFUInt32
{
	FF_EFFECT = 0,
	FF_SOURCE = 1,
	FF_MIXER  = 2
};

// Plugin capabilities
// Old enumerators that are no longer in use are tagged with comments ("No longer in use").
// This is to discourage using their values for new enumerators.
enum class PluginCapability : FFUInt32
{
	FF_CAP_16BITVIDEO         = 0,// No longer in use
	FF_CAP_24BITVIDEO         = 1,// No longer in use
	FF_CAP_32BITVIDEO         = 2,// No longer in use
	FF_CAP_PROCESSFRAMECOPY   = 3,// No longer in use
	FF_CAP_PROCESSOPENGL      = 4,// No longer in use
	FF_CAP_SETTIME            = 5,
	FF_CAP_MINIMUMINPUTFRAMES = 10,
	FF_CAP_MAXIMUMINPUTFRAMES = 11,
	FF_CAP_COPYORINPLACE      = 15,// No longer in use
};

// Old return values that were used for the FF_CAP_COPYORINPLACE capability.
enum class CopyOrInPlacePreference : FFUInt32
{
	FF_CAP_PREFER_NONE    = 0,// No longer in use
	FF_CAP_PREFER_INPLACE = 1,// No longer in use
	FF_CAP_PREFER_COPY    = 2,// No longer in use
	FF_CAP_PREFER_BOTH    = 3,// No longer in use
};

// Parameter types
enum class ParameterType : FFUInt32
{
	FF_TYPE_BOOLEAN    = 0,
	FF_TYPE_EVENT      = 1,
	FF_TYPE_RED        = 2,
	FF_TYPE_GREEN      = 3,
	FF_TYPE_BLUE       = 4,
	FF_TYPE_XPOS       = 5,
	FF_TYPE_YPOS       = 6,
	FF_TYPE_STANDARD   = 10,
	FF_TYPE_OPTION     = 11,
	FF_TYPE_BUFFER     = 12,
	FF_TYPE_INTEGER    = 13,
	FF_TYPE_TEXT       = 100,
	FF_TYPE_HUE        = 200,
	FF_TYPE_SATURATION = 201,
	FF_TYPE_BRIGHTNESS = 202,
	FF_TYPE_ALPHA      = 203,
	FF_FAIL            = SuccessFail::FF_FAIL
};

// Input status
enum class InputStatus : FFUInt32
{
	FF_INPUT_NOTINUSE = 0,
	FF_INPUT_INUSE    = 1,
	FF_FAIL           = SuccessFail::FF_FAIL

};

// Parameter usages
enum class ParameterUsage : FFUInt32
{
	FF_USAGE_STANDARD = 0,
	FF_USAGE_FFT      = 1,
	FF_FAIL           = SuccessFail::FF_FAIL
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FreeFrame Types
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Forward declaration of types referenced in union FFMixed
class PluginBase;
typedef struct PluginInfoStructTag PluginInfoStruct;
typedef struct PluginExtendedInfoStructTag PluginExtendedInfoStruct;
typedef struct SetParameterStructTag SetParameterStruct;
typedef struct SetBeatinfoStructTag SetBeatinfoStruct;
typedef struct SetHostinfoStructTag SetHostinfoStruct;
typedef struct RangeStructTag RangeStruct;
typedef struct GetRangeStructTag GetRangeStruct;
typedef struct ViewportStructTag ViewportStruct;
typedef struct TextureStructTag TextureStruct;
typedef struct ProcessOpenGLStructTag ProcessOpenGLStruct;
typedef struct GetParameterElementNameStructTag GetParameterElementNameStruct;
typedef struct GetParameterElementValueStructTag GetParameterElementValueStruct;
typedef struct SetParameterElementValueStructTag SetParameterElementValueStruct;

// FFMixed is used to interchange values between the plugin and host in a single 4-byte buffer
// It can be used to hold an integer (including various enums), float, or pointer (to a string, struct, class object
// 2019-07-11 ffgl@subpixels.com
// - Extend FFMixed union value types to make pointer casting legible and maintainable
typedef union FFMixed
{
	// Basic numeric types
	FFUInt32 UIntValue;
	float FloatValue;

	// Generic pointer (avoid using this; try to use pointers to specific classes/structs)
	void* PointerValue;

	// Doubles don't fit, so use a pointer to a double
	double* pDouble;

	// C-style string pointer
	const char* StrValue;

	// Enumerations
	InputStatus InputStatus;
	ParameterType ParameterType;
	ParameterUsage ParameterUsage;
	PluginCapability PluginCapability;

	// Struct / class object pointers
	PluginBase* pFreeFrameGLPlugin;
	const PluginInfoStruct* pPluinInfo;
	const PluginExtendedInfoStruct* pPluginExtendedInfo;
	SetParameterStruct* pSetParameter;
	SetBeatinfoStruct* pSetBeatinfo;
	SetHostinfoStruct* pSetHostinfo;
	RangeStruct* pRange;
	GetRangeStruct* pGetRange;
	ViewportStruct* pViewport;
	TextureStruct* pTexture;
	ProcessOpenGLStruct* pProcessOpenGL;
	GetParameterElementNameStruct* pGetParameterElementName;
	GetParameterElementValueStruct* pGetParameterElementValue;
	SetParameterElementValueStruct* pSetParameterElementValue;
} FFMixed;

typedef void* FFInstanceID;

// PluginInfoStruct
typedef struct PluginInfoStructTag
{
	FFUInt32 APIMajorVersion;
	FFUInt32 APIMinorVersion;
	char PluginUniqueID[ 4 ];// 4 chars uniqueID - not null terminated
	char PluginName[ 16 ];   // 16 chars plugin friendly name - not null terminated
	PluginType PluginType; // Effect or source
} PluginInfoStruct;

// PluginExtendedInfoStruct
typedef struct PluginExtendedInfoStructTag
{
	FFUInt32 PluginMajorVersion;
	FFUInt32 PluginMinorVersion;
	char* Description;
	char* About;
	FFUInt32 FreeFrameExtendedDataSize;
	void* FreeFrameExtendedDataBlock;
} PluginExtendedInfoStruct;

// SetParameterStruct
typedef struct SetParameterStructTag
{
	FFUInt32 ParameterNumber;
	FFMixed NewParameterValue;
} SetParameterStruct;

// SetBeatinfoStruct
// 2019-07-12 ffgl@subpixels.com - changed bpm and barPhrase types from FFMixed to float
typedef struct SetBeatinfoStructTag
{
	float bpm;
	float barPhase;
} SetBeatinfoStruct;

// SetHostinfoStruct
typedef struct SetHostinfoStructTag
{
	const char* name;
	const char* version;
} SetHostinfoStruct;

typedef struct RangeStructTag
{
	float min;
	float max;
} RangeStruct;

typedef struct GetRangeStructTag
{
	FFUInt32 parameterNumber;
	RangeStruct range;
} GetRangeStruct;

//ViewportStruct (for InstantiateGL)
typedef struct ViewportStructTag
{
	GLuint x, y, width, height;
} ViewportStruct;

//TextureStruct (for ProcessOpenGLStruct)
typedef struct TextureStructTag
{
	FFUInt32 Width, Height;
	FFUInt32 HardwareWidth, HardwareHeight;
	GLuint Handle;//the actual texture handle, from glGenTextures()
} TextureStruct;

// ProcessOpenGLStruct
typedef struct ProcessOpenGLStructTag
{
	FFUInt32 numInputTextures;
	TextureStruct** inputTextures;

	//if the host calls ProcessOpenGL with a framebuffer object actively bound
	//(as is the case when the host is capturing the plugins output to an offscreen texture)
	//the host must provide the GL handle to its EXT_framebuffer_object
	//so that the plugin can restore that binding if the plugin
	//makes use of its own FBO's for intermediate rendering
	GLuint HostFBO;
} ProcessOpenGLStruct;

// GetParameterElementNameStruct
typedef struct GetParameterElementNameStructTag
{
	FFUInt32 ParameterNumber;
	FFUInt32 ElementNumber;
} GetParameterElementNameStruct;

// GetParameterElementValueStruct
typedef struct GetParameterElementValueStructTag
{
	FFUInt32 ParameterNumber;
	FFUInt32 ElementNumber;
} GetParameterElementValueStruct;

// SetParameterElementValueStruct
typedef struct SetParameterElementValueStructTag
{
	FFUInt32 ParameterNumber;
	FFUInt32 ElementNumber;
	FFMixed NewParameterValue;
} SetParameterElementValueStruct;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Function prototypes
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// plugMain - The one and only exposed function
// parameters:
//  functionCode - tells the plugin which function is being called
//  inputValue - 32-bit parameter or architecture specific pointer to parameter structure (eg 64 bits on x64)
//  instanceID - the id of the instance on which you want to invoke the functionCode
//
// PLUGIN DEVELOPERS:  you shouldn't need to change this function
//
// All parameters are cast as untyped pointers and cast to appropriate
// types here
//
// All return values are cast to untyped pointers here before return to
// the host
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifdef _WIN32

BOOL APIENTRY DllMain( HANDLE hModule, DWORD ul_reason_for_call, LPVOID lpReserved );

__declspec( dllexport ) FFMixed __stdcall plugMain( FFUInt32 functionCode, FFMixed inputValue, FFInstanceID instanceID );
typedef __declspec( dllimport ) FFMixed( __stdcall* FF_Main_FuncPtr )( FFUInt32, FFMixed, FFInstanceID );

#else

//linux and Mac OSX share these
FFMixed plugMain( FFUInt32 functionCode, FFMixed inputValue, FFInstanceID instanceID );
typedef FFMixed ( *FF_Main_FuncPtr )( FFUInt32 funcCode, FFMixed inputVal, FFInstanceID instanceID );

#endif

}//End namespace ffgl

#ifndef _WIN32
}//extern "C"
#endif
