//
// Copyright (c) 2004 - InfoMus Lab - DIST - University of Genova
//
// InfoMus Lab (Laboratorio di Informatica Musicale)
// DIST - University of Genova
//
// http://www.infomus.dist.unige.it
// news://infomus.dist.unige.it
// mailto:staff@infomus.dist.unige.it
//
// Developer: Gualtiero Volpe
// mailto:volpe@infomus.dist.unige.it
//
// Developer: Trey Harrison
// www.harrisondigitalmedia.com
//
// Last modified: October 26 2006
//
// 2019-07-22 ffgl@subpixels.com
// - use enum ParameterType
// - use enum ParameterUsage

#pragma once

#include "FFGL.h"
#include "ParamBase.h"
#include <vector>
#include <string>

namespace ffgl
{

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// \class		PluginManager
///	\brief		PluginManager manages information concerning a plugin inputs, parameters, and capabilities.
/// \authors	Gualtiero Volpe, Menno Vink (menno@resolume.com)
/// \version	2.0.0
///
/// The PluginManager class is the base class for FreeFrameGL plugins developed with the FreeFrameGL SDK since it provides
/// them with methods for automatically manage information concerning plugin inputs, paramaters, and capabilities.
/// Examples of information managed by this class are the number of inputs and parameters of a plugin; the name, type and
/// default value of each parameter.
/// Plugins developed with the FFGL SDK (and thus having this class as base class) should call the protected methods
/// of this class in order to specify the information related to their inputs and parameters. These calls
/// are usually done while constructing the plugin subclass. Plugins subclasses should also call methods of this class in
/// order to get information about the images they are going to process (i.e., their width and height).
/// The default implementations of the FFGL global functions call the public methods of this class in order to
/// return to the host information about a plugin.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class PluginManager
{
public:
	/// The standard destructor of PluginManager.
	virtual ~PluginManager();

	/// This method returns the minimum number of inputs the host must provide.
	/// It is usually called by the default implementations of the FreeFrame global functions.
	///
	/// \return		The minimum number of inputs the host must provide.
	FFUInt32 GetMinInputs() const;

	/// This method returns the maximum number of inputs the plugin can receive.
	/// It is usually called by the default implementations of the FreeFrame global functions.
	///
	/// \return		The maximum number of inputs the plugin can receive.
	FFUInt32 GetMaxInputs() const;

	/// This function allows you to add a new parameter to the plugin. There is differents kind of
	/// parameters available, Adding a parameter allows the plugin to be
	/// aware of them, make them available to the host and take of all the communication with it.
	/// It can also allow the plugin to automatically pass the current value of each parameter to the
	/// shader before drawing (see ffglqs::Plugin::SendParams()).
	/// \param	param		The parameter to add
	void AddParam( ParamBase* param );

	/// Find Param (for editing)
	/// \param  index  The index of the parameter you want to edit
	/// \return        Pointer to the param object; nullptr if index is out of range
	ParamBase* FindParam( FFUInt32 index );

	/// Find Param (for editing), cast to a specific parameter type
	/// \param  ParamT  A class that inherits from ParamBase (eg ParamFloat)
	/// \param  index  The index of the parameter you want to edit
	/// \return        Pointer to the param object, cast to a pointer to the template class;
	///                nullptr if index is out of range or the object is not of the specified class
	template< class ParamT >
	ParamT* FindParam( FFUInt32 index );

	// Get Param (read only)
	/// \param  index  The index of the parameter you want to access (note: const)
	/// \return        Pointer to the param object; nullptr if index is out of range
	const ParamBase* GetParam( FFUInt32 index ) const;

	/// Find Param (for editing), cast to a specific parameter type
	/// \param  ParamT  A class that inherits from ParamBase (eg ParamFloat)
	/// \param  index  The index of the parameter you want to access (note: const)
	/// \return        Pointer to the param object, cast to a pointer to the template class;
	///                nullptr if index is out of range or the object is not of the specified class
	template< class ParamT >
	const ParamT* GetParam( FFUInt32 index ) const;

	/// This method returns how may parameters the plugin has.
	/// It is usually called by the default implementations of the FreeFrame global functions.
	///
	/// \return		The number of parameters of the plugin.
	FFUInt32 GetNumParams() const;

	/// This method returns the name of the plugin parameter whose index is passed as parameter
	/// to the method. It is usually called by the default implementations of the FreeFrame global functions.
	///
	/// \param	index	    The index of the plugin parameter whose name is queried.
	///						It should be in the range [0, Number of plugin parameters).
	/// \return				The name of the plugin parameter whose index is passed to the method.
	///						The return value is a pointer to an array of 16 1-byte ASCII characters,
	///						not null terminated (see FreeFrame specification). NULL is returned on error.
	const char* GetParamName( FFUInt32 index ) const;

	/// This method is called to know the type of the plugin parameter whose index is passed as parameter
	/// to the method. It is usually called by the default implementations of the FreeFrame global functions.
	///
	/// \param	paramIndex	The index of the plugin parameter whose type is queried.
	///						It should be in the range [0, Number of plugin parameters).
	/// \return				The type of the plugin parameter whose index is passed as parameter to the method.
	///						Codes for allowed parameter types are defined in FFGL.h.
	///						In case of error, FF_FAIL is returned.
	ParameterType GetParamType( FFUInt32 index ) const;

	/// This method is called to know the number of elements of the plugin parameter whose index is passed as parameter
	/// to the method. It is usually called by the default implementations of the FreeFrame global functions.
	/// By default parameters will return 1.
	/// In case parameter is an array, it will return the number of elements in the array.
	/// \param	index	    The index of the plugin parameter whose number of elements is being queried.
	///						It should be in the range [0, Number of plugin parameters).
	/// \return				The type of the plugin parameter whose index is passed as parameter to the method.
	///						Codes for allowed parameter types are defined in FreeFrame.h.
	///						In case of error, FF_FAIL is returned.
	FFResult GetNumParamElements( FFUInt32 index ) const;

	/// Get the name of an element for a parameter that can have elements (PluginType::FF_TYPE_OPTION)
	/// The host may use this name to display that element on the screen. For example an OPTION's element
	/// will most likely be show as a dropdown option, which when pressed will cause the param's value to change
	/// to that element's default value.
	///
	/// \param	paramIndex	The index of the plugin parameter which contains the element for which
	///						the name is being queried.
	///						It should be in the range[0, Number of plugin parameters).
	///	\param	elemIndex	The index of the element whose name is queried.
	///						It should be in the range [0, parameter.Number of elements).
	///	\return				The name of the parameter's element whose indices are passed to the method.
	///						The return value is a pointer to a null terminated c string of ASCII characters.
	///						In case of error, NULL is returned.
	const char* GetParamElementName( FFUInt32 paramIndex, FFUInt32 elemIndex ) const;

	/// Get the default value of an element for a parameter whose elements default values have any effect (eg ParameterType::FF_TYPE_OPTION)
	/// The host will use this value to set the param to this specific value when that option is chosen.
	///
	/// \param	paramIndex	The index of the plugin parameter which contains the element for which
	///						the default value is being queried.
	///						It should be in the range[0, Number of plugin parameters).
	///	\param	elemIndex	The index of the element whose default value is queried.
	///						It should be in the range [0, parameter.Number of elements).
	///	\return				The default value of the parameter's element whose indices are passed to the method.
	///						The return value should be cast either to a float*.
	///						In case of error, NULL is returned.
	FFMixed GetParamElementDefault( FFUInt32 paramIndex, FFUInt32 elemIndex ) const;

	/// Set the value of an element for a parameter whose elements values can change (eg ParameterType::FF_TYPE_BUFFER)
	/// The host will use this to change the contents of a parameter's elements, which the plugin can then use
	/// depending on what the buffer is used for.
	///
	/// \param	paramIndex	The index of the plugin parameter which contains the element for which
	///						the value is to be changed.
	///						It should be in the range[0, Number of plugin parameters).
	///	\param	elemIndex	The index of the element whose value is to be  changed.
	///						It should be in the range [0, parameter.Number of elements).
	///	\return				FFGL result indicating if setting the value succeeded. Setting a value might fail
	///						if either of the provided indices is out of range. FF_SUCCESS on success, FF_FAIL otherwise.
	FFResult SetParamElementValue( FFUInt32 paramIndex, FFUInt32 elemIndex, float value );

	/// This method is called to know the usage of a plugin parameter
	/// For example a float buffer that expects a spectrum, will return FF_USAGE_FFT
	/// By default parameters will return ParameterUsage::FF_USAGE_STANDARD.
	/// In case parameter is an array, it will return the number of elements in the array.
	/// \param	paramIndex	The index of the plugin parameter whose name is queried.
	///						It should be in the range [0, Number of plugin parameters).
	/// \return				The usage of the plugin parameter whose index is passed as parameter to the method.
	///						Codes for allowed parameter types are defined in FreeFrame.h.
	///						In case of error, FF_FAIL is returned.
	ParameterUsage GetParamUsage( FFUInt32 index ) const;

	/// This method is called to get the default value of the plugin parameter whose index is passed as parameter
	/// to the method. It is usually called by the default implementations of the FreeFrame global functions.
	///
	/// \param	paramIndex	The index of the plugin parameter whose name is queried.
	///						It should be in the range [0, Number of plugin parameters).
	/// \return				The default value of the plugin parameter whose index is passed as parameter to the method.
	///						The return value should be cast either to a char* in case of text parameters or to a float*
	///						in any other case. In case of error, NULL is returned.
	FFMixed GetParamDefault( FFUInt32 dwIndex ) const;

	/// Reset a parameter to its default value
	///
	/// \param	paramIndex	The index of the plugin parameter whose name is queried.
	///						It should be in the range [0, Number of plugin parameters).
	FFResult ResetParam( FFUInt32 index );

	/// This method is called by a the host to determine whether the plugin supports the SetTime function
	bool GetTimeSupported() const;

	RangeStruct GetParamRange( FFUInt32 index ) const;

protected:
	// Plugin parameters. ParamBase is used for polymorphism.
	std::vector< ParamBase* > params;

	///	The standard constructor of PluginManager.
	/// \remark	Notice that the PluginManager constructor is a protected member function, i.e., nor PluginManager
	///			objects nor CFreeFramePlugin objects should be created directly, but only objects of the subclasses
	///			implementing specific plugins should be instantiated.
	PluginManager();

	/// This method is called by a plugin subclass, derived from this class, to indicate the minimum number
	/// of inputs the host must provide. This method is usually called when a plugin object is instantiated
	/// (i.e., in the plugin subclass constructor).
	///
	/// \param	iMinInputs	The plugin subclass should set it to the minimum number of inputs
	///						the host must provide.
	void SetMinInputs( FFUInt32 iMinInputs );

	/// This method is called by a plugin subclass, derived from this class, to indicate the maximum number
	/// of inputs the plugin can receive. This method is usually called when a plugin object is instantiated
	/// (i.e., in the plugin subclass constructor).
	///
	/// \param	iMaxInputs	The plugin subclass should set it to the maximum number of inputs the plugin
	///						can receive.
	void SetMaxInputs( FFUInt32 iMaxInputs );

	/// This method is called by a plugin subclass, derived from this class, to indicate whether the
	/// SetTime function is supported
	///
	/// \param	supported	The plugin indicates whether it supports the SetTime function by passing true or false (1 or 0)
	void SetTimeSupported( bool supported );

	void SetParamRange(FFUInt32 index, float min, float max);

private:
	// Inputs
	FFUInt32 m_iMinInputs;
	FFUInt32 m_iMaxInputs;

	// Time capability
	bool m_timeSupported;
};

}//End namespace ffgl
